# +++ Documentation fbdraw +++

Sa 29. Apr 20:28:51 CEST 2023

The purpose of these tools is to create framebuffer graphics in userspace
in an easy and scritable way.

# compile

Just type "make".
The source code is optimized for use with musl-gcc.
CC=musl-gcc is therefore also set in the Makefile.

# Examples

make
make test
make demo

# Documentation

## fbenv

this tools opens the framebuffer device, reads resolution and color depth,
and sets the following environment variables

"SCREEN_HEIGHT"
"SCREEN_WIDTH"
"SCREEN_COLORDEPTH"

existing values will be overwritten.
fbenv forks a child process with this changed environment.

Example calls:

fbenv env | grepSCREEN
fbenv drawfb < demodata.txt
fbenv sh

W=$(fbenv sh -c 'echo "$SCREEN_WIDTH"')

## fbdraw

Draws to the framebuffer.

The tool reads commands and parameters line by line from stdin and
also clones this line unchanged to stdout. It will only
interprets the first 63 characters of a line. The remaining
Characters per line are discarded.

See demodata.txt for examples

The following commands exist:

cmd	arguments	name	description
-------------------------------------------------- ---------------------------------
c	r g b 		color 	Sets the character color to R G B
m	x y 		move	Moves the drawing cursor to desired coordinates
r	dx dy 		rect	Draws a rectangle
p	x y r g b	pixel	Draws exactly one pixel at x y in color r g b
s	n		sleep	Waits specified number of microseconds
o	ox oy		offset	Defines global offset for drawing operations

Commands and parameters/arguments are separated by a space.
All arguments MUST be integers.

Examples:

c 140 20 0		Switches to the color red
m 20 80			Changes to coordinates x=20 and y=80
r 10 10			Draws a 10x10 px square
r 10 40			Draws a rectangle 10px wide and 40px high
s 50000			Waits 50 milliseconds (blocking!) (used for animations)
p 23 42 164 163 255 	Draws a light blue pixel x=23 y=42
o 50 100		Defines a global offset of x=50 y=100
 
Example call: fbdraw < demodata.txt > /dev/null

Comment lines can start with a # and are ignored.
All unknown commands are also ignored.

fbdraw is parameterized via *environment variables*
 
/dev/fb0 is used as the framebuffer device by default,
as long as the corresponding environment variable "FBDEV" is not set.
You can use a different framebuffer device by

export FBDEV=/dev/fb1

sets, or also directly before calling fbdraw

FBDEV=/dev/fb1 fbdraw < demodata.txt > /dev/null

The screen resolution is also set via environment variables like this:

export SCREEN_WIDTH=1920
export SCREEN_HEIGHT=1080
export SCREEN_COLORDEPTH=32

With the clone function, you can also output to multiple devices
draw and use different resolution at the same time


cat demodata.txt | \
FBDEV=/dev/fb0 fbenv fbdraw | \
FBDEV=/dev/fb1 SCREEN_COLORDEPTH=16 SCREEN_WIDTH=1024 SCREEN_HEIGHT=768 fbdraw | \
FBDEV=/dev/fb2 SCREEN_COLORDEPTH=32 SCREEN_WIDTH=800 SCREEN_HEIGHT=480 fbdraw | \
> /dev/null

fbdraw discards drawing statements outside of each
screen area.

## fbmouse

Prints the cursor coordinates of the mouse to stdout.
In addition, the events "Left click", "Right click", "Middle click"
and the scroll wheel movements "up" and "down" recorded and as
Event issued. Each event is preceded by the "#" character,
so that it can be filtered very well in scripts. In addition, at events
also takes a 64 px grid raster into account and calls it the tile id.
See the further examples.

fbmouse is parameterized via *environment variables*

export SCREEN_WIDTH=1920
export SCREEN_HEIGHT=1080
export MOUSEDEVICE=/dev/input/mice

The screen size is important to generate the correct Tile ID
and also to define the maximum values.

You can also use fbenv here to set the environment variables
Set "SCREEN_WIDTH" and "SCREEN_HEIGHT" automatically

MOUSEDEVICE=/dev/input/mice fbenv fbmouse

Events: click_left, click_right, click_middle, scroll_up, scroll_down

Output format on click/scroll: "# event tile_id x y"
Output format when moving the mouse: "x y" (without leading hash)


## fbtouch

Similar to fbmouse, but for touch screen events, since these events
always return *absolute coordinates* instead of *relative coordinates*
and are queried completely differently.

The parameterization of fbtouch is done via *environment variables*

export SCREEN_WIDTH=800
export SCREEN_HEIGHT=400
export TOUCHDEVICE=/dev/input/event0

Note that you have to choose the right event device!
You can also select manufacturer and model name.
Please refer:

ls -1 /dev/input/by-id/

The screen size is important to generate the correct Tile ID
and also to define the maximum values.

You can also use fbenv here to set the environment variables
Set "SCREEN_WIDTH" and "SCREEN_HEIGHT" automatically

MOUSEDEVICE=/dev/input/mice fbenv fbmouse

Events: click_left, click_right, click_middle, scroll_up, scroll_down

Output format on click/scroll: "# event tile_id x y"
Output format when moving the mouse: "x y" (without leading hash)


## fbtouch

Similar to fbmouse, but for touch screen events, the
always return *absolute coordinates* instead of *relative coordinates*
and be queried completely differently.

The parameterization of fbtouch is done via *environment variables*

export SCREEN_WIDTH=800
export SCREEN_HEIGHT=400
export TOUCHDEVICE=/dev/input/event0

optional:  export MULTITOUCH=1

Note that you have to choose the right event device!
You can also select manufacturer and model name.
Please refer:

ls -1 /dev/input/by-id/

The screen size is important to generate the correct Tile ID
and also to define the maximum values.

You can also use fbenv here to set the environment variables
Set "SCREEN_WIDTH" and "SCREEN_HEIGHT" automatically.
So an example call to fbtouch looks like this:

TOUCHDEVICE=/dev/input/event0 fbenv fbtouch

If you have a multitouch device and want to use conversion mode
to singletouch, you have to set a environment variable called "MULTITOUCH".

MULTITOUCH=1 TOUCHDEVICE=/dev/input/event0 fbenv fbtouch

Output format: "# touch tile_id x y"

# Further examples

fbenv fbmouse | grep "#" | cut -d " " -f 3

only shows events (e.g. clicks or scroll wheel events)
and mentions the tile ID in each case. the tile ID refers
on your grid of 64 pixels, i.e. a virtual fixed grid
of 64x64 pixels, which is laid over the screen area.

Example at a resolution of 1920x1080
30 horizontal blocks divided by 64:
(1920/64)+1=30+1=31 .. we start at ID 1 instead of ID 0

```
[ 1][ 2][ 3][ 4][ 5]...[31]
[32][33]...............[60]
. .
......................[512]
```
So to see if the grid on the tile with ID 128
was clicked with the left mouse button, you only have to
use something like this:

fbenv fbmouse | stdbuf -i0 -o0 grep -q "# click_left 128" && echo ok

fbenv fbtouch | stdbuf -i0 -o0 grep -q "# touch 1" && echo "top left corner click detected"


# Tips and Tricks

## How can I convert my own pixel graphics into the fbdraw p draw format?

Use the shell script "convert2fbdraw.sh" in the subfolder "examples".
It uses imagemagick to convert known graphic formats into rgb values
(top left to bottom right) line by line to "fbdraw p-formatted" lines
on stdout. The script also requires xxd and gawk.

Hint: Try to use transparent background to save disk space and to
speed up the drawing process. So only draw the pixels you really need.
To do this, simply choose a fix background color, e.g. black "0 0 0",
then omit these p commands with grep:

cat example.jpg | convert2fbdraw.sh 640 | grep -v "^p .* 0 0 0$" | gzip -9 -c > result.txt.gz


## I want to quickly draw a background graphic in the framebuffer. How do I do that?

Since line-by-line drawing with p commands is quite slow, you should
first of all draw the image completely into the framebuffer.
Then you save the complete framebuffer of your device in a
compressed file. Then you can use the framebuffer again at any time
initially overwrite before you apply further fbdraw commands to it again.

Example:

Let's assume you have a background of 800 by 480 pixels in
the file background.jpg. Now we draw this into the framebuffer
and then save them in background.fb.gz. We can overwrite the framebuffer
again with zcat anytime and without using fbdraw at all.
However, keep in mind, that the background image is "baked"
for this device then, since it uses a fixed depth and resolution!!

convert2fbdraw.sh 800 < background.jpg | fbenv fbdraw

cat /dev/fb0 | gzip -9 -c > background_800x480.fb.gz
zcat background_800x480.fb.gz > /dev/fb0
