/*

reinhard@finalmedia.de
Do 16. Mär 15:04:28 CET 2023

draw colored rectangles to framebuffer

read commands from stdin line by line
and cloning lines to stdout
lines longer than 64 chars will get truncated
(details see demodata.txt)

command is single char
all arguments are integer only 0-9
all elements are space separated

cmd	arguments	name	description
-------------------------------------------
c	r g b         	color	set color
m	x y           	move	move to coord x y
o	ox oy          	offset	set global offset x y
r	dx dy         	rect	draw rectangle with given dimensions
p	x y r g b     	pixel	draw pixel in given color
s	n             	sleep	sleep n microseconds

example usage:
-----------------------
export FBDEV=/dev/fb0
export SCREEN_HEIGHT=768
export SCREEN_WIDTH=1024
export SCREEN_COLORDEPTH=32
fbdraw << EOF > /dev/null
c 230 70 20
m 45 50
r 128 128
c 250 90 40
m 55 60
r 108 108
EOF

*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdint.h>
#include <unistd.h>

#define	RGB16 (r >> 3) << 11 | (g >> 2) << 5 | b >> 3
#define	RGB32 r << 16 | g << 8 | b

int main () {

  //default env if not set
  setenv("SCREEN_WIDTH","1920",0);
  setenv("SCREEN_HEIGHT","1080",0);
  setenv("SCREEN_COLORDEPTH","32",0);
  setenv("FBDEV","/dev/fb0",0);

  int width = atoi(getenv("SCREEN_WIDTH"));
  int height = atoi(getenv("SCREEN_HEIGHT"));
  int colordepth = atoi(getenv("SCREEN_COLORDEPTH"));

  const int buffer_width = width;
  const int buffer_height = height;
  const int buffer_size = buffer_width * buffer_height * (colordepth/8);
  uint32_t (*mem)[width];

  char cmd[64];
  int arg1, arg2, arg3, arg4, arg5;
  int x = 0;
  int y = 0;
  int mx = 0;
  int my = 0;
  int ox = 0;	// global offset x
  int oy = 0;	// global offset y
  int r = 255;
  int g = 255;
  int b = 255;

  int fd = open(getenv("FBDEV"), O_RDWR);
  mem = (uint32_t (*)[buffer_width]) mmap(NULL, buffer_size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);


  while (1) {
    if (feof(stdin)) break;

    // max command length per line is 63 chars
    fgets(cmd, 63, stdin);
    // clone to stdout
    fputs(cmd,stdout);

    if (sscanf(cmd, "s %d\n", &arg1) == 1) {	// sleep given microseconds
      usleep(arg1);
    } else if (sscanf(cmd, "c %d %d %d\n", &arg1, &arg2, &arg3) == 3) {	// set rgb color
      r = arg1;
      g = arg2;
      b = arg3;
    } else if (sscanf(cmd, "m %d %d\n", &arg1, &arg2) == 2) {		// move to
      x = arg1;
      y = arg2;
    } else if (sscanf(cmd, "o %d %d\n", &arg1, &arg2) == 2) {		// set global offset
      ox = arg1;
      oy = arg2;
    } else if (sscanf(cmd, "p %d %d %d %d %d\n", &arg1, &arg2, &arg3, &arg4, &arg5) == 5) {	// draw single pixel
      x = arg1;
      y = arg2;
      r = arg3;
      g = arg4;
      b = arg5;
      mx = ox+x;
      my = oy+y;
	if (mx<buffer_width && my<buffer_height) {
		if (colordepth<32) {
			mem[my][mx] = RGB16;
		} else {
			mem[my][mx] = RGB32;
		}
	}
    } else if (sscanf(cmd, "r %d %d\n", &arg1, &arg2) == 2) {		// draw rectangle
      for (int dx = 0; dx < arg1; dx++) {
        for (int dy = 0; dy < arg2; dy++) {
		mx = ox+x+dx;
		my = oy+y+dy;
        	if (mx<buffer_width && my<buffer_height) {
			if (colordepth<32) {
				mem[my][mx] = RGB16;
			} else {
				mem[my][mx] = RGB32;
			};
		}
        }
      }
    }
  }

  if (!feof(stdin)) fclose(stdin);
  munmap(mem, buffer_size);

}
